#import "LNSQTSkimmerView.h"
#import "LNSQTMovie.h"
#import "NSBezierPath+PXRoundedRectangleAdditions.h"
#import "NSImage-LNSAdditions.h"


//
//	Theory of operation:
//
//	The view is divided into a series of "cells" (not NSCell instances) that are organized into
//	rows.  Each "cell" displays a poster frame for the time range it represents.  The "cell"
//	containing the "currentTime" displays the currentTime's frame rather than the first frame
//	of th cells time range.
//
//
//
//	Performance Notes
//
//	The class makes use of QTMovie's frameImageAtTime: which is actually quite a slow call.  In
//	particular, if you are viewing a movie where QT must calculate frames from a nearby Keyframe
//	the call can be very slow.  My tests show that the Apple Intermediate codec is a good one
//	to use as it seems to produce frames fairly quickly.  I have gone to some length to reduce 
//	the number of frames drawn to reduce the load on the frameImageAtTime: call.  Still, there
//	is more to do to get scrolling and live window resize to perform acceptably.
//


#define kHeaderHeight				30.0
#define kFooterHeight				5.0
#define kClipHeight					96.0
#define kClipTopSpacing				5.0
#define kClipBottomSpacing			10.0
#define kClipLeftSpacing			15.0
#define kClipRightSpacing			15.0
#define kRightHeight				(kClipTopSpacing + (kClipHeight * mThumbnailScale) + kClipBottomSpacing)
#define kTimestampHSpacing			2.0
#define kIntercellSpacing			1.0
#define kSelectionTimestampHSpacing 4.0
#define kTimestampShadowWidth		4.0
#define kSelectionFrameWidth		3.0
#define kSelectionFrameInset		1.5
#define kSelectionShadowWidth		4.0
#define kSelectionColor				[NSColor colorWithCalibratedRed:249.0 / 255.0 green:217.0 / 255.0 blue:26.0 / 255.0 alpha:1.0]
#define kInactiveSelectionColor		[NSColor lightGrayColor]
#define kCurrentTimeTextColor		[NSColor whiteColor]
#define kCurrentTimeBarColor		[NSColor redColor]
#define kCurrentTimeBarBackgroundColor \
									[NSColor whiteColor]
#define kTimeFont					[NSFont boldSystemFontOfSize:9.5]
#define kBackgroundColor			[NSColor colorWithCalibratedWhite:26.0 / 255.0 alpha:1.0]
#define kUnselectedHighlightColor	[NSColor colorWithDeviceWhite:0.1 alpha:0.3]

#define USE_FRAME_CACHING
//#define DEBUG_LOGGING
//#define DEBUG_DRAWING


@interface LNSQTSkimmerView (Private)

- (void) _setupView;
- (void) _configureForMovie;
- (void) _timeDidChange:(NSNotification*) notification;
- (void) _selectionDidChange:(NSNotification*) notification;
- (NSImage*) _cachedImageAtTime:(QTTime) time;
- (NSImage*) _frameImageAtTime:(QTTime) time canQueue:(BOOL) canQueue;
- (void) _refreshCellForTime:(QTTime) time;
- (void) _refreshCellsForTimeRange:(QTTimeRange) timeRange;
- (void) _refreshCellsFrom:(unsigned) startCellIndex to:(unsigned) endCellIndex;
- (unsigned) _cellIndexForTime:(QTTime) time;
- (unsigned) _cellIndexForPoint:(NSPoint) point strict:(BOOL) strict;
- (QTTime) _timeForCellAtIndex:(unsigned) index;
- (QTTime) _timeForPoint:(NSPoint) point;
- (NSRect) _frameForTime:(QTTime) time;
- (NSRect) _frameForCellAtIndex:(unsigned) index;
- (NSRect) _refreshFrameForCellAtIndex:(unsigned) index;
- (BOOL) _movieIsPlaying;
- (NSString*) _stringForTime:(QTTime) time;
- (NSDictionary*) _selectionTextAttributes;
- (NSDictionary*) _currentTimeTextAttributes;
- (void) _adjustSelectionStartForTime:(QTTime) time;
- (void) _adjustSelectionForTime:(QTTime) time;
- (NSShadow*) _textShadow;
- (NSShadow*) _selectionShadow;
- (NSImage*) _selectionGrabber;
- (QTTime) _pinTime:(QTTime) time InRange:(QTTimeRange) range;
- (BOOL) _isZeroTime:(QTTime) time;
- (BOOL) _isZeroTimeRange:(QTTimeRange) range;

@end


@implementation LNSQTSkimmerView (Private)

- (void) _setupView
{
	mThumbnailScale = 1.0;
	mDesiredCellDuration = QTMakeTimeWithTimeInterval(5.0);
	mCachedImages = [[NSMutableDictionary dictionaryWithCapacity:100] retain];
	mImagesToLoad = [[NSMutableArray arrayWithCapacity:100] retain];
}

- (void) _configureForMovie
{
	static BOOL inConfigureForMovie = NO;
	
	if (!inConfigureForMovie)
	{
		NSScrollView* scroller = [self enclosingScrollView];
		inConfigureForMovie = YES;
		
		float boundsWidth = scroller ? [scroller contentSize].width : [self frame].size.width;
		NSSize movieSize = mMovie ? [[mMovie attributeForKey:QTMovieNaturalSizeAttribute] sizeValue] : NSZeroSize;
		float scale =  (mThumbnailScale * kClipHeight) / movieSize.height;
		float rowHeight = kRightHeight;
		NSTimeInterval cellDuration;

		mMovieDuration = mMovie ? [[mMovie attributeForKey:QTMovieDurationAttribute] QTTimeValue] : QTZeroTime;
		mScaledMovieSize.width = floorf(scale * movieSize.width + 0.5);
		mScaledMovieSize.height = floorf(scale * movieSize.height + 0.5);

		QTGetTimeInterval(mMovieDuration, &mMovieDurationInterval);
		QTGetTimeInterval(mDesiredCellDuration, &cellDuration);

		mCells = mMovieDurationInterval == 0.0 ? 0 : ceilf(mMovieDurationInterval / cellDuration);
		mCellsPerRow = mCells == 0 ? 0 : ((boundsWidth - kClipLeftSpacing) - kClipRightSpacing) / (mScaledMovieSize.width + kIntercellSpacing);
		if (mCellsPerRow < 1)
			mCellsPerRow = 1;
		mRows = ceilf((float) mCells / (float) mCellsPerRow);
		mCellDuration = mCells > 0 ? QTMakeTimeWithTimeInterval(mMovieDurationInterval / mCells) : mDesiredCellDuration;
		mMovieFrameCount = [[[[[mMovie tracksOfMediaType:QTMediaTypeVideo] lastObject] media] attributeForKey:QTMediaSampleCountAttribute] longValue];

		QTGetTimeInterval(mCellDuration, &mCellDurationInterval);
		
		[super setFrameSize:NSMakeSize(boundsWidth, ((float) mRows) * rowHeight + kHeaderHeight + kFooterHeight)];
		[self setNeedsDisplay:YES];
		
		inConfigureForMovie = NO;
		
		//	If the scroll view auto-hides scrollers, the change in height may have had the side-effect of
		//	changing the width.  When this happens, we have to reconfigure again to properly account for
		//	the change in width.
		if (boundsWidth != [self frame].size.width)
		{
			[NSObject cancelPreviousPerformRequestsWithTarget:self
													 selector:@selector(_configureForMovie)
													   object:nil];
			[self performSelector:@selector(_configureForMovie) withObject:nil afterDelay:0.01];
		}
	}
}

- (void) _timeDidChange:(NSNotification*) notification
{
	if (mDragKind == kLNSQTScrubDrag_None)
	{
		QTTime time = [mMovie currentTime];
		QTMovie* saveMovie = mMovie;

		mMovie = nil;	// don't allow LNSQTSkimmerView to try and update the movie's currentTime since we are
						// being called from the movie
		[self setCurrentTime:time];
		mMovie = saveMovie;
		if ([self _movieIsPlaying])
			[self scrollRectToVisible:[self _frameForTime:time]];
	}
}

- (void) _selectionDidChange:(NSNotification*) notification
{
	if (mDragKind == kLNSQTScrubDrag_None)
	{
		QTMovie* saveMovie = mMovie;
		
		mMovie = nil;	// don't allow LNSQTSkimmerView to try and update the movie's selection since we are
						// being called from the movie
		[self setSelection:QTMakeTimeRange([saveMovie selectionStart], [saveMovie selectionDuration])];
		mMovie = saveMovie;
	}
}

- (void) _fetchImageIntoCache:(NSNumber*) key
{
	QTTime time = QTMakeTimeWithTimeInterval([key doubleValue]);
	unsigned frameIndex = [key doubleValue] / (mMovieDurationInterval / mMovieFrameCount);

	key = [NSNumber numberWithUnsignedLong:frameIndex];
		
	if (![mCachedImages objectForKey:key])
	{
		//[mCachedImages setObject:[[mMovie frameImageAtTime:time] scaledImageWithSize:mScaledMovieSize]
		[mCachedImages setObject:[mMovie frameImageAtTime:time]
						  forKey:key];
		[self _refreshCellForTime:time];
	}
}

- (void) _fetchImageIntoCache
{
	//	QTMovie's frameImageAtTime: is apparently *not* totally thread safe.  To avoid this problem, I use
	//	idle processing to perform the differed loading of movie frames.  The _frameImageAtTime method queues
	//	images for loading.  _fetchImageIntoCache is invoked periodically and loads images into the cache and
	//	then marks the frame's cell is needing display.
	//
	//	The problem with this appraoch is that background frame loading only occurs when the app is idle.  If
	//	the user is dragging a slider, tracking a menu, etc., then frame loading is blocked.

	for (int i = 5; i > 0 && [mImagesToLoad count] > 0; --i)
	{
		NSNumber* key = [mImagesToLoad objectAtIndex:0];
		QTTime time = QTMakeTimeWithTimeInterval([key doubleValue]);
		unsigned frameIndex = [key doubleValue] / (mMovieDurationInterval / mMovieFrameCount);
		
		key = [NSNumber numberWithUnsignedLong:frameIndex];
		
		[mImagesToLoad removeObjectAtIndex:0];
		if (![mCachedImages objectForKey:key])
		{
//			[mCachedImages setObject:[[mMovie frameImageAtTime:time] scaledImageWithSize:mScaledMovieSize]
			[mCachedImages setObject:[mMovie frameImageAtTime:time]
							  forKey:key];
			[self _refreshCellForTime:time];
		}
	}

	if ([mImagesToLoad count] > 0)
		[self performSelector:@selector(_fetchImageIntoCache) withObject:nil afterDelay:0.001];
}

- (NSImage*) _cachedImageAtTime:(QTTime) time
{
#ifdef USE_FRAME_CACHING
	NSTimeInterval interval;
	
	QTGetTimeInterval(time, &interval);

	unsigned frameIndex = interval / (mMovieDurationInterval / mMovieFrameCount);
	return [mCachedImages objectForKey:[NSNumber numberWithUnsignedLong:frameIndex]];
#else
	return nil;
#endif
}

- (NSImage*) _frameImageAtTime:(QTTime) time canQueue:(BOOL) canQueue
{
#ifdef USE_FRAME_CACHING
	NSTimeInterval interval;

	QTGetTimeInterval(time, &interval);
	
	unsigned frameIndex = interval / (mMovieDurationInterval / mMovieFrameCount) + (5.0 / time.timeScale);
	NSNumber* key = [NSNumber numberWithUnsignedLong:frameIndex];
	NSImage* result = [mCachedImages objectForKey:key];
	
	if (!result)
	{
		if (canQueue)
		{
#if 1
			[mImagesToLoad addObject:[NSNumber numberWithDouble:interval]];

			[NSObject cancelPreviousPerformRequestsWithTarget:self
													 selector:@selector(_fetchImageIntoCache)
													   object:nil];
			[self performSelector:@selector(_fetchImageIntoCache) withObject:nil afterDelay:0.01];
#else
			[NSObject cancelPreviousPerformRequestsWithTarget:self
													 selector:@selector(_fetchImageIntoCache:)
													   object:[NSNumber numberWithDouble:interval]];
			[self performSelector:@selector(_fetchImageIntoCache:) withObject:key afterDelay:0.01];
#endif
		}
		else
		{
			result = [mMovie frameImageAtTime:time];
			if (result)
				[mCachedImages setObject:result forKey:key];
		}
	}
	return result;
#else
	return [mMovie frameImageAtTime:time];
#endif
}

- (void) _refreshCellForTime:(QTTime) time
{
	[self setNeedsDisplayInRect:[self _refreshFrameForCellAtIndex:[self _cellIndexForTime:time]]];
}

- (void) _refreshCellsForTimeRange:(QTTimeRange) timeRange
{
	unsigned startIndex = [self _cellIndexForTime:timeRange.time];
	unsigned endIndex = [self _cellIndexForTime:QTTimeRangeEnd(timeRange)];
	
	if (startIndex == NSNotFound)
		startIndex = 0;
	if (endIndex == NSNotFound)
		endIndex = startIndex;
		
	[self _refreshCellsFrom:startIndex to:endIndex];
}

- (void) _refreshCellsFrom:(unsigned) startCellIndex to:(unsigned) endCellIndex
{
	NSAssert(startCellIndex != NSNotFound, @"startCellIndex is NSNotFound");
	NSAssert(endCellIndex != NSNotFound, @"endCellIndex is NSNotFound");
	
	if (startCellIndex > endCellIndex)
	{
		unsigned tmp = startCellIndex;

		startCellIndex = endCellIndex;
		endCellIndex = tmp;
	}
	
	unsigned selStartRow = startCellIndex / mCellsPerRow;
	unsigned selEndRow = endCellIndex / mCellsPerRow;

	if (selStartRow == selEndRow)
		[self setNeedsDisplayInRect:NSUnionRect([self _refreshFrameForCellAtIndex:startCellIndex],
												[self _refreshFrameForCellAtIndex:endCellIndex])];
	else
	{
		[self setNeedsDisplayInRect:NSUnionRect([self _refreshFrameForCellAtIndex:startCellIndex],
												[self _refreshFrameForCellAtIndex:selStartRow * mCellsPerRow + mCellsPerRow - 1])];

		for (unsigned row = selStartRow + 1; row < selEndRow; ++row)
			[self setNeedsDisplayInRect:NSUnionRect([self _refreshFrameForCellAtIndex:row * mCellsPerRow],
													[self _refreshFrameForCellAtIndex:row * mCellsPerRow + mCellsPerRow - 1])];
		
		[self setNeedsDisplayInRect:NSUnionRect([self _refreshFrameForCellAtIndex:selEndRow * mCellsPerRow],
												[self _refreshFrameForCellAtIndex:endCellIndex])];
	}
}

- (unsigned) _cellIndexForTime:(QTTime) time
{
	NSTimeInterval interval;

	QTGetTimeInterval(time, &interval);

	NSTimeInterval result = interval / mCellDurationInterval + ((double) 5.0 / (double) time.timeScale);

#ifdef DEBUG_LOGGING
	NSLog(@"LNSQTSkimmerView _cellIndexForTime: %@[%f] (%@[%f]): %d", 
				QTStringFromTime(time), interval, QTStringFromTime(mCellDuration), cellDuration, (unsigned) result);
#endif

	return result;
}

- (unsigned) _cellIndexForPoint:(NSPoint) point strict:(BOOL) strict
{
	if (strict)
	{
		if (point.x < kClipLeftSpacing)
			return NSNotFound;
		if (point.y < kHeaderHeight)
			return NSNotFound;
	}
	
	float rowHeight = kRightHeight;
	unsigned row = floorf((point.y - kHeaderHeight) / rowHeight);
	unsigned col = (point.x - kClipLeftSpacing) / (mScaledMovieSize.width + kIntercellSpacing);

	if (col >= mCellsPerRow)
	{
		if (strict)
			return NSNotFound;
		else
			col = mCellsPerRow - 1;
	}
	
	unsigned index = row * mCellsPerRow + col;

	if (index >= mCells)
		return strict ? NSNotFound : mCells - 1;
	else
		return index;
}

- (QTTime) _timeForCellAtIndex:(unsigned) index
{
	NSParameterAssert(index != NSNotFound);
		
	return QTMakeTimeWithTimeInterval(mCellDurationInterval * index);
}

- (QTTime) _timeForPoint:(NSPoint) point
{
	unsigned cellIndex = [self _cellIndexForPoint:point strict:NO];

	if (cellIndex == NSNotFound)
		return QTZeroTime;
	else
	{
		NSRect cellFrame = [self _frameForCellAtIndex:cellIndex];
		QTTime cellStart = [self _timeForCellAtIndex:cellIndex];

		if (point.x > NSMaxX(cellFrame))
			point.x = NSMaxX(cellFrame) - 0.5;

		double pos = point.x - NSMinX(cellFrame);
		if (pos >= 0.0)
		{
			NSTimeInterval cellInterval = mCellDurationInterval * (pos / NSWidth(cellFrame));
			
			return QTTimeIncrement(cellStart, QTMakeTimeWithTimeInterval(cellInterval));
		}
		else
			return QTZeroTime;
	}
}

- (NSRect) _frameForTime:(QTTime) time
{
	unsigned cellIndex = [self _cellIndexForTime:time];
	NSRect cellFrame = [self _frameForCellAtIndex:cellIndex];
	QTTime cellStartTime = [self _timeForCellAtIndex:cellIndex];
	NSTimeInterval cellStartInterval;
	NSTimeInterval timeInterval;
	
	QTGetTimeInterval(cellStartTime, &cellStartInterval);
	QTGetTimeInterval(time, &timeInterval);

	double xPos = floor(((timeInterval - cellStartInterval) / mCellDurationInterval) * (mScaledMovieSize.width + kIntercellSpacing) + 0.5);
	return NSMakeRect(NSMinX(cellFrame) + xPos, NSMinY(cellFrame), 1.0, NSHeight(cellFrame));
}

- (NSRect) _frameForCellAtIndex:(unsigned) index
{
	NSParameterAssert(index != NSNotFound);
	float rowHeight = kRightHeight;
	unsigned row = index / mCellsPerRow;
	unsigned col = index - (row * mCellsPerRow);

	return NSIntegralRect(NSMakeRect(kClipLeftSpacing + col * (mScaledMovieSize.width + kIntercellSpacing),
							  row * rowHeight + kClipTopSpacing + kHeaderHeight,
							  mScaledMovieSize.width,
							  mScaledMovieSize.height));
}

- (NSRect) _refreshFrameForCellAtIndex:(unsigned) index
{
	//	The refresh frame may be different that the cell frame due to things like current time
	//	indicators and other decorations that may fall outside the cell's frame
	NSRect frame = [self _frameForCellAtIndex:index];

	//	If the cell contains the current frame, we need to expand the frame to encompase the
	//	space needed to draw the current time string
	if (index == [self _cellIndexForTime:mCurrentTime])
	{
		NSRect r = [self _frameForTime:mCurrentTime];
		
		// account for the current frame bar
		r = NSInsetRect(r, -1.0, 0.0);
		// account for the current frame timestamp
		r.size.width += [[self _stringForTime:mCurrentTime] sizeWithAttributes:[self _currentTimeTextAttributes]].width +
						kTimestampHSpacing + (kTimestampShadowWidth - kTimestampHSpacing);

		frame = NSUnionRect(frame, r);
	}
	
	//	If the selection is not empty, then we must account for it...
	if (![self _isZeroTimeRange:mSelection] || mHighlightSelection)
	{
		QTTime selectionEnd = QTTimeRangeEnd(mSelection);
		unsigned selStartCell = [self _cellIndexForTime:mSelection.time];
		unsigned selEndCell = [self _cellIndexForTime:selectionEnd];

		//	If the cell contains the selection start, we need to allow for the added width of the selection frame and the knob
		if (index == selStartCell)
		{
			float delta = [[self _selectionGrabber] size].width / 2.0 + kSelectionShadowWidth + kSelectionFrameInset;
			NSRect r = NSInsetRect([self _frameForCellAtIndex:index], -delta, -delta);
			
			frame = NSUnionRect(frame, r);
		}
		//	if the cell contains the selection end, we need to allow for the added width of the selection frame and the knob
		//	*and* the selection duration string
		else if (index == selEndCell)
		{
			float delta = [[self _selectionGrabber] size].width / 2.0 + kSelectionShadowWidth + kSelectionFrameInset;
			NSRect r = NSInsetRect([self _frameForCellAtIndex:index], -delta, -delta);

			if (mDragKind != kLNSQTScrubDrag_None)
			{
				NSRect r1 = [self _frameForTime:selectionEnd];

				r1.size.width += [[self _stringForTime:mSelection.duration] sizeWithAttributes:[self _currentTimeTextAttributes]].width +
									kSelectionTimestampHSpacing + (kTimestampShadowWidth - kSelectionTimestampHSpacing);
				r = NSUnionRect(r, r1);
			}
			frame = NSUnionRect(frame, r);
		}
		//	If the cell lies within the range of selected cells, expand it to include the selection frame
		else if (index > selStartCell && index < selEndCell)
		{
			float delta = kSelectionShadowWidth + kSelectionFrameInset;
			NSRect r = NSInsetRect([self _frameForCellAtIndex:index], -delta, -delta);

			frame = NSUnionRect(frame, r);
		}
	}

#ifdef DEBUG_LOGGING
	NSLog(@"LNSQTSkimmerView _refreshFrameForCellAtIndex: %d (%@): %@", 
			index, NSStringFromRect([self _frameForCellAtIndex:index]), NSStringFromRect(frame));
#endif

	return frame;
}

- (BOOL) _movieIsPlaying { return mMovie ? [mMovie rate] != 0 : NO; }

- (NSString*) _stringForTime:(QTTime) time
{
	static NSCharacterSet* charSet = nil;
	NSString* s = QTStringFromTime(time);

	//	Trim QuickTime's timescale stuff off the end of the string...
	if (!charSet)
		charSet = [[NSCharacterSet characterSetWithCharactersInString:@"0123456789:"] retain];
	[[NSScanner scannerWithString:s] scanCharactersFromSet:charSet intoString:&s];
	
	//	If the duration is seconds, add the 's' unit to the end of the string
	if ([s hasPrefix:@"0:00:00:"])
		s = [NSString stringWithFormat:@"%@s", [s substringWithRange:NSMakeRange(8, [s length] - 8)]];

	//	Trim leading 0's and :'s...
	unsigned pos = 0;
	unsigned len = [s length];
	unichar ch = [s characterAtIndex:pos];
	while ((ch == '0' || ch == ':') && ++pos < len - 2)
		ch = ch = [s characterAtIndex:pos];

	return [s substringWithRange:NSMakeRange(pos, len - pos)];
}

- (NSDictionary*) _selectionTextAttributes
{
	static NSDictionary* sTextAttrs = nil;
	
	 if (!sTextAttrs)
		sTextAttrs = [[NSDictionary dictionaryWithObjectsAndKeys:
												kTimeFont, NSFontAttributeName,
												kSelectionColor, NSForegroundColorAttributeName,
												nil] retain];

	return sTextAttrs;
}

- (NSDictionary*) _currentTimeTextAttributes
{
	static NSDictionary* sTextAttrs = nil;
	
	 if (!sTextAttrs)
		sTextAttrs = [[NSDictionary dictionaryWithObjectsAndKeys:
												kTimeFont, NSFontAttributeName,
												kCurrentTimeTextColor, NSForegroundColorAttributeName,
												nil] retain];

	return sTextAttrs;
}

- (void) _adjustSelectionStartForTime:(QTTime) time
{
	//	We need to adjust the start time to be the time in the current selection furthest away from the
	//	time the user just clicked on to make the selection extend naturally.
	NSTimeInterval timeInterval;
	NSTimeInterval selStartInterval;
	NSTimeInterval selEndInterval;

	QTGetTimeInterval(time, &timeInterval);
	QTGetTimeInterval(mSelection.time, &selStartInterval);
	QTGetTimeInterval(QTTimeRangeEnd(mSelection), &selEndInterval);
	
	if (timeInterval < selStartInterval) // user clicked before current selection
		mSelectionStart = QTTimeRangeEnd(mSelection);
	else if (timeInterval > selEndInterval) // user clicked after current selection
		mSelectionStart = mSelection.time;
	else // user clicked within current selection
	{
		//	We need to figure out if they cleared nearer the start (meaning that they want to
		//	alter the beginning of the selection) or nearer the end of the selection (meaning 
		//	they want to alter the end of the selection)
		
		if (timeInterval < selStartInterval + (selEndInterval - selStartInterval) / 2.0)
			mSelectionStart = QTTimeRangeEnd(mSelection);
		else
			mSelectionStart = mSelection.time;
	}
}

- (void) _adjustSelectionForTime:(QTTime) time
{
	switch (QTTimeCompare(mSelectionStart, time))
	{
	case NSOrderedAscending:
		[self setSelection:QTMakeTimeRange(mSelectionStart, QTTimeDecrement(time, mSelectionStart))];
		break;

	case NSOrderedDescending:
		[self setSelection:QTMakeTimeRange(time, QTTimeDecrement(mSelectionStart, time))];
		break;
	
	default:
		[self setSelection:QTMakeTimeRange(mSelectionStart, QTZeroTime)];
		break;
	}
}

- (NSShadow*) _textShadow;
{
	static NSShadow* sShadow = nil;
	
	if (!sShadow)
	{
		sShadow = [[NSShadow alloc] init];
		[sShadow setShadowOffset:NSZeroSize];
		[sShadow setShadowBlurRadius:kTimestampShadowWidth];
		[sShadow setShadowColor:[NSColor blackColor]];
	}
	
	return sShadow;
}

- (NSShadow*) _selectionShadow;
{
	static NSShadow* sShadow = nil;
	
	if (!sShadow)
	{
		sShadow = [[NSShadow alloc] init];
		[sShadow setShadowOffset:NSZeroSize];
		[sShadow setShadowBlurRadius:kSelectionShadowWidth];
		[sShadow setShadowColor:[NSColor blackColor]];
	}
	
	return sShadow;
}

- (NSImage*) _selectionGrabber
{
	static NSImage* sImage = nil;
	
	if (!sImage)
		sImage = [[NSImage imageNamed:@"grabber"] retain];
	NSAssert(sImage, @"failed to load selection grammer resource");
	return sImage;
}

- (QTTime) _pinTime:(QTTime) time InRange:(QTTimeRange) range
{
	QTTime rangeEnd = QTTimeRangeEnd(range);
	
	if (QTTimeCompare(time, rangeEnd) == NSOrderedDescending)
		time = rangeEnd;
	else if (QTTimeCompare(time, range.time) == NSOrderedAscending)
		time = range.time;
	return time;
}

- (BOOL) _isZeroTime:(QTTime) time { return QTTimeCompare(time, QTZeroTime) == NSOrderedSame; }

- (BOOL) _isZeroTimeRange:(QTTimeRange) range { return [self _isZeroTime:range.duration]; }

@end


@implementation LNSQTSkimmerView

- (id)initWithFrame:(NSRect)frameRect
{
	if ((self = [super initWithFrame:frameRect]) != nil)
	{
		[self _setupView];
	}

	return self;
}

- (id)initWithCoder:(NSCoder*) coder
{
	if ((self = [super initWithCoder:coder]))
	{
		[self _setupView];

		mThumbnailScale = [coder decodeFloatForKey:@"thumbnailScale"];
		mDesiredCellDuration = [coder decodeQTTimeForKey:@"cellDuration"];
		mHighlightSelection = [coder decodeBoolForKey:@"highlightSelection"];
		[self setMovie:[coder decodeObjectForKey:@"movie"]];
	}
	
	return self;
}

- (void) encodeWithCoder:(NSCoder*) coder
{
	[super encodeWithCoder:coder];

	[coder encodeFloat:mThumbnailScale forKey:@"thumbnailScale"];
	[coder encodeQTTime:mDesiredCellDuration forKey:@"cellDuration"];
	[coder encodeObject:mMovie forKey:@"movie"];
	[coder encodeBool:mHighlightSelection forKey:@"highlightSelection"];
}

- (void) dealloc
{
	[NSObject cancelPreviousPerformRequestsWithTarget:self];
	[self setMovie:nil];
	[mCachedImages release];
	[mImagesToLoad release];
	[super dealloc];
}

- (BOOL) isFlipped { return YES; }
- (BOOL) acceptsFirstResponder { return YES; }
- (BOOL) acceptsFirstMouse:(NSEvent*) theEvent { return YES; }

- (BOOL) becomeFirstResponder
{
	[self _refreshCellsForTimeRange:mSelection];
	return [super becomeFirstResponder];
}

- (BOOL) resignFirstResponder
{
	[self _refreshCellsForTimeRange:mSelection];
	return [super resignFirstResponder];
}

- (void) drawRect:(NSRect) rect
{
#ifdef DEBUG_DRAWING
	[[NSColor greenColor] set];
	NSFrameRect([self bounds]);
#endif

	if (mCells > 0)
	{
		BOOL hasSelection = mHighlightSelection || ![self _isZeroTimeRange:mSelection];
		QTTime selectionEnd = QTTimeRangeEnd(mSelection);
		unsigned firstCell = [self _cellIndexForPoint:NSMakePoint(NSMinX(rect) + 1.0, NSMinY(rect) + 1.0) strict:NO];
		unsigned lastCell = [self _cellIndexForPoint:NSMakePoint(NSMaxX(rect) - 1.0, NSMaxY(rect) - 1.0) strict:NO];
		unsigned currentCell = [self _cellIndexForTime:mCurrentTime];
		unsigned selStartCell = hasSelection ? [self _cellIndexForTime:mSelection.time] : NSNotFound;
		unsigned selEndCell = hasSelection ? [self _cellIndexForTime:selectionEnd] : NSNotFound;

		if (lastCell == NSNotFound)
			lastCell = mCells - 1;

#ifdef DEBUG_LOGGING
		NSLog(@"LNSQTSkimmerView: drawRect: %@, firstCell: %d, lastCell: %d", NSStringFromRect(rect), firstCell, lastCell);
#endif

		//	Draw the cells that have been dirtied
		[NSGraphicsContext saveGraphicsState];
		{
			[[NSGraphicsContext currentContext] setImageInterpolation:NSImageInterpolationHigh];
			for (unsigned index = firstCell; index <= lastCell; ++index)
			{
				NSRect area = [self _frameForCellAtIndex:index];
				
				if ([self needsToDrawRect:area])
				{
					NSImage* movieFrame = nil;
					
					if (currentCell == index && ![self _movieIsPlaying])
					{
						movieFrame = [self _cachedImageAtTime:mCurrentTime];
						if (!movieFrame)
							movieFrame = [mMovie frameImageAtTime:mCurrentTime];
					}
					else if (selStartCell == index && selStartCell != selEndCell)
					{
						movieFrame = [self _cachedImageAtTime:mSelection.time];
						if (!movieFrame)
							movieFrame = [mMovie frameImageAtTime:mSelection.time];
					}
					else if (selEndCell == index && selStartCell != selEndCell)
					{
						movieFrame = [self _cachedImageAtTime:selectionEnd];
						if (!movieFrame)
							movieFrame = [mMovie frameImageAtTime:selectionEnd];
					}
					else
						movieFrame = [self _frameImageAtTime:[self _timeForCellAtIndex:index] canQueue:YES];

					NSSize movieFrameSize = [movieFrame size];

					[movieFrame setFlipped:YES];
					[movieFrame drawInRect:area
								  fromRect:NSMakeRect(0.0, 0.0, movieFrameSize.width, movieFrameSize.height)
								 operation:NSCompositeSourceOver
								  fraction:1.0];
					[movieFrame setFlipped:NO];

					if (mHighlightSelection && (index < selStartCell || index > selEndCell))
					{
						[kUnselectedHighlightColor set];
						NSRectFillUsingOperation(area, NSCompositePlusDarker);
					}
					
					[[NSColor blackColor] set];
					NSFrameRect(NSInsetRect(area, -1.0, -1.0));
				}
#ifdef DEBUG_LOGGING
				else
					NSLog(@"LNSQTSkimmerView: drawRect: skipped cell %d", index);
#endif
			}
		}
		[NSGraphicsContext restoreGraphicsState];

		//	Draw current frame indicator
		if (mDragKind == kLNSQTScrubDrag_None && ![self _isZeroTime:mCurrentTime] && currentCell >= firstCell && currentCell <= lastCell)
		{
			NSRect r = [self _frameForTime:mCurrentTime];
			
			[kCurrentTimeBarBackgroundColor set];
			NSRectFill(NSInsetRect(r, -1.0, 0.0));
			[kCurrentTimeBarColor set];
			NSFrameRect(r);
			
			NSString* s = [self _stringForTime:mCurrentTime];
			NSDictionary* sAtts = [self _currentTimeTextAttributes];
			NSSize sSize = [s sizeWithAttributes:sAtts];
			
			[NSGraphicsContext saveGraphicsState];
			{
				[[self _textShadow] set];
				[s drawInRect:NSMakeRect(NSMaxX(r) + kTimestampHSpacing, 
										 NSMinY(r) + (NSHeight(r) - sSize.height) / 2.0, 
										 sSize.width, sSize.height) withAttributes:sAtts];
			}
			[NSGraphicsContext restoreGraphicsState];
		}

		//	Draw selection
		if (hasSelection)
		{
			//	The selection may span a number of rows of cells.
			BOOL isFirstResponder = [[self window] firstResponder] == self && [[self window] isKeyWindow];
			NSImage* grabber = [self _selectionGrabber];
			NSSize grabberSize = [grabber size];
			
			//	We need to iterate over the rows that intersect the selected cells
			unsigned selStartRow = selStartCell / mCellsPerRow;
			unsigned selEndRow = selEndCell / mCellsPerRow;

			if (selStartRow == selEndRow)
			{
				NSRect r = NSUnionRect([self _frameForTime:mSelection.time],
									   [self _frameForTime:selectionEnd]);
				NSBezierPath* path = [NSBezierPath bezierPathWithRoundedRect:NSInsetRect(r, -kSelectionFrameInset, -kSelectionFrameInset) 
																cornerRadius:6.0];

				if (mHighlightSelection)
				{
					NSRect leftR = [self _frameForCellAtIndex:selStartCell];
					NSRect rightR = [self _frameForCellAtIndex:selEndCell];
										   
					[kUnselectedHighlightColor set];
					NSRectFillUsingOperation(NSMakeRect(NSMinX(leftR),
														NSMinY(leftR),
														NSMinX(r) - NSMinX(leftR), 
														NSHeight(r)), NSCompositePlusDarker);
					NSRectFillUsingOperation(NSMakeRect(NSMaxX(r),
														NSMinY(r),
														NSMaxX(rightR) - NSMaxX(r), 
														NSHeight(r)), NSCompositePlusDarker);
				}

				[path setLineWidth:kSelectionFrameWidth];
				[NSGraphicsContext saveGraphicsState];
				{
					[[self _selectionShadow] set];
					[(isFirstResponder ? kSelectionColor : kInactiveSelectionColor) set];
					[path stroke];
				}
				[NSGraphicsContext restoreGraphicsState];

				if (isFirstResponder)
				{
					[grabber drawInRect:NSMakeRect(NSMinX(r) - grabberSize.width / 2.0 - kSelectionFrameInset,
												   NSMinY(r) + (NSHeight(r) - grabberSize.height) / 2.0,
												   grabberSize.width, grabberSize.height)
							   fromRect:NSMakeRect(0.0, 0.0, grabberSize.width, grabberSize.height)
							  operation:NSCompositeSourceOver
							   fraction:1.0];
					[grabber drawInRect:NSMakeRect(NSMaxX(r) - grabberSize.width / 2.0 + kSelectionFrameInset,
												   NSMinY(r) + (NSHeight(r) - grabberSize.height) / 2.0,
												   grabberSize.width, grabberSize.height)
							   fromRect:NSMakeRect(0.0, 0.0, grabberSize.width, grabberSize.height)
							  operation:NSCompositeSourceOver
							   fraction:1.0];
				}
			}
			else
			{
				NSRect r2 = [self _frameForCellAtIndex:selStartRow * mCellsPerRow + mCellsPerRow - 1];
				NSRect r = NSUnionRect([self _frameForTime:mSelection.time],
									   NSMakeRect(NSMaxX(r2), NSMinY(r2), 1.0, NSHeight(r2)));
				NSBezierPath* path = [NSBezierPath bezierPathWithRoundedRect:NSInsetRect(r, -kSelectionFrameInset, -kSelectionFrameInset) 
																cornerRadius:6.0
																   inCorners:OSTopLeftCorner | OSBottomLeftCorner];

				if (mHighlightSelection)
				{
					NSRect leftR = [self _frameForCellAtIndex:selStartCell];

					[kUnselectedHighlightColor set];
					NSRectFillUsingOperation(NSMakeRect(NSMinX(leftR),
														NSMinY(leftR),
														NSMinX(r) - NSMinX(leftR), 
														NSHeight(r)), NSCompositePlusDarker);
				}

				[path setLineWidth:kSelectionFrameWidth];
				[NSGraphicsContext saveGraphicsState];
				{
					[[self _selectionShadow] set];
					[(isFirstResponder ? kSelectionColor : kInactiveSelectionColor) set];
					[path stroke];
				}
				[NSGraphicsContext restoreGraphicsState];

				if (isFirstResponder)
					[grabber drawInRect:NSMakeRect(NSMinX(r) - grabberSize.width / 2.0 - kSelectionFrameInset,
												   NSMinY(r) + (NSHeight(r) - grabberSize.height) / 2.0,
												   grabberSize.width, grabberSize.height)
							   fromRect:NSMakeRect(0.0, 0.0, grabberSize.width, grabberSize.height)
							  operation:NSCompositeSourceOver
							   fraction:1.0];

				for (unsigned row = selStartRow + 1; row < selEndRow; ++row)
				{
					NSRect r = NSUnionRect([self _frameForCellAtIndex:row * mCellsPerRow],
										   [self _frameForCellAtIndex:row * mCellsPerRow + mCellsPerRow - 1]);
					NSBezierPath* path = [NSBezierPath bezierPathWithRect:NSInsetRect(r, -kSelectionFrameInset, -kSelectionFrameInset)];

					[path setLineWidth:kSelectionFrameWidth];
					[NSGraphicsContext saveGraphicsState];
					{
						[[self _selectionShadow] set];
						[(isFirstResponder ? kSelectionColor : kInactiveSelectionColor) set];
						[path stroke];
					}
					[NSGraphicsContext restoreGraphicsState];
				}
				
				r2 = [self _frameForCellAtIndex:selEndRow * mCellsPerRow];
				r = NSUnionRect(NSMakeRect(NSMinX(r2), NSMinY(r2), 1.0, NSHeight(r2)),
								[self _frameForTime:selectionEnd]);
				path = [NSBezierPath bezierPathWithRoundedRect:NSInsetRect(r, -kSelectionFrameInset, -kSelectionFrameInset) 
												  cornerRadius:6.0
													 inCorners:OSTopRightCorner | OSBottomRightCorner];

				if (mHighlightSelection)
				{
					NSRect rightR = [self _frameForCellAtIndex:selEndCell];

					[kUnselectedHighlightColor set];
					NSRectFillUsingOperation(NSMakeRect(NSMaxX(r),
														NSMinY(r),
														NSMaxX(rightR) - NSMaxX(r), 
														NSHeight(r)), NSCompositePlusDarker);
				}

				[path setLineWidth:kSelectionFrameWidth];
				[NSGraphicsContext saveGraphicsState];
				{
					[[self _selectionShadow] set];
					[(isFirstResponder ? kSelectionColor : kInactiveSelectionColor) set];
					[path stroke];
				}
				[NSGraphicsContext restoreGraphicsState];

				if (isFirstResponder)
					[grabber drawInRect:NSMakeRect(NSMaxX(r) - grabberSize.width / 2.0 + kSelectionFrameInset,
												   NSMinY(r) + (NSHeight(r) - grabberSize.height) / 2.0,
												   grabberSize.width, grabberSize.height)
							   fromRect:NSMakeRect(0.0, 0.0, grabberSize.width, grabberSize.height)
							  operation:NSCompositeSourceOver
							   fraction:1.0];
			}

			if (mDragKind == kLNSQTScrubDrag_AlteringSelection)
			{
				//	Draw the selection duration time...
				NSString* s = [self _stringForTime:mSelection.duration];
				NSDictionary* sAtts = [self _selectionTextAttributes];
				NSSize sSize = [s sizeWithAttributes:sAtts];
				NSRect r = [self _frameForTime:selectionEnd];

				[NSGraphicsContext saveGraphicsState];
				{
					[[self _textShadow] set];
					[s drawInRect:NSMakeRect(NSMaxX(r) + kSelectionTimestampHSpacing, 
											 NSMinY(r) + (NSHeight(r) - sSize.height) / 2.0, 
											 sSize.width, sSize.height) withAttributes:sAtts];
					
				}
				[NSGraphicsContext restoreGraphicsState];
			}
		}

		//	Draw clip duration
		if (firstCell == 0)
		{
			NSRect r = [self _frameForTime:QTZeroTime];
			NSString* s = [self _stringForTime:mMovieDuration];
			NSDictionary* sAtts = [self _currentTimeTextAttributes];
			NSSize sSize = [s sizeWithAttributes:sAtts];
			
			[NSGraphicsContext saveGraphicsState];
			{
				[[self _textShadow] set];
				[s drawInRect:NSMakeRect(NSMaxX(r) + kTimestampHSpacing, 
										 NSMinY(r) + (NSHeight(r) - sSize.height) / 2.0, 
										 sSize.width, sSize.height) withAttributes:sAtts];
				
			}
			[NSGraphicsContext restoreGraphicsState];
		}
	}
}

- (void) viewDidMoveToSuperview
{
	if ([self superview])
	{
		NSScrollView* scroller = [self enclosingScrollView];
		NSRect frame = [self frame];
		
		frame.size.width = [scroller contentSize].width;
		[self setAutoresizingMask:NSViewWidthSizable | NSViewMaxYMargin];
		[self setFrame:frame];
		
		[scroller setBackgroundColor:kBackgroundColor];
		[scroller setDrawsBackground:YES];
	}
}

- (void) viewDidMoveToWindow
{
	[[self window] setAcceptsMouseMovedEvents:YES];
}

- (QTMovie*) movie { return [[mMovie retain] autorelease]; }
- (void) setMovie:(QTMovie*) movie
{
	if (mMovie != movie)
	{
		NSNotificationCenter* notifier = [NSNotificationCenter defaultCenter];

		[notifier removeObserver:self
							name:QTMovieTimeDidChangeNotification 
						  object:mMovie];
		[notifier removeObserver:self
							name:LNSQTMoviePlayingTimeDidChangeNotification 
						  object:mMovie];
		[notifier removeObserver:self
							name:QTMovieSelectionDidChangeNotification 
						  object:mMovie];

		[self willChangeValueForKey:@"currentTime"];
		[self willChangeValueForKey:@"selection"];

		[mMovie release];
		mMovie = [movie retain];
		[mCachedImages removeAllObjects];
		[mImagesToLoad removeAllObjects];
		
		if (mMovie)
		{
			mSelectionStart = [mMovie selectionStart];
			mSelection = QTMakeTimeRange(mSelectionStart, [mMovie selectionDuration]);
			mCurrentTime = [mMovie currentTime];

			[notifier addObserver:self 
						 selector:@selector(_timeDidChange:) 
							 name:QTMovieTimeDidChangeNotification 
						   object:mMovie];
			[notifier addObserver:self 
						 selector:@selector(_timeDidChange:) 
							 name:LNSQTMoviePlayingTimeDidChangeNotification 
						   object:mMovie];
			[notifier addObserver:self 
						 selector:@selector(_selectionDidChange:) 
							 name:QTMovieSelectionDidChangeNotification 
						   object:mMovie];
		}
		else
		{
			mSelection = QTMakeTimeRange(QTZeroTime, QTZeroTime);
			mCurrentTime = QTZeroTime;
		}

		[self didChangeValueForKey:@"selection"];
		[self didChangeValueForKey:@"currentTime"];

		[self _configureForMovie];
	}
}

- (float) thumbnailScale { return mThumbnailScale; }
- (void) setThumbnailScale:(float) newValue
{
	if (mThumbnailScale != newValue)
	{
		mThumbnailScale = newValue;
		[self _configureForMovie];
		[self scrollRectToVisible:[self _frameForTime:mCurrentTime]];
	}
}

- (QTTime) cellDuration
{
	if (QTTimeCompare(mDesiredCellDuration, QTMakeTimeWithTimeInterval(999999.0)) == NSOrderedSame)
		return mMovieDuration;
	else
		return mDesiredCellDuration;
}
- (void) setCellDuration:(QTTime) time
{
	if (QTTimeCompare(time, mCellDuration) != NSOrderedSame) 
	{
		//	mDesiredCellDuration represents the user visible cell duration setting.  However, this value
		//	generally does not divide evenly into the movie duration.  So, _configureForMovie calculates
		//	the real cell duration that results in the frames being evenly direibuted among the cells.

		mDesiredCellDuration = time;
		[self _configureForMovie];
		[self scrollRectToVisible:[self _frameForTime:mCurrentTime]];
	}
} 

- (unsigned) cellDurationStep
{
	NSTimeInterval cellDurationInterval;

	QTGetTimeInterval(mDesiredCellDuration, &cellDurationInterval);
	
	switch ((unsigned) cellDurationInterval)
	{
	default:
		return 0;

	case 1:
		return 1;

	case 2:
		return 2;

	case 5:
		return 3;

	case 10:
		return 4;

	case 30:
		return 5;

	case 999999:
		return 6;
	}
}
- (void) setCellDurationStep:(unsigned) step
{
	//	The UI for setting the cell duration uses a non-liniar scale.  I suppose I could have used some
	//	sort of value transformer to convert the 0-6 scale of the slider into duration times (secons),
	//	but this seemed more expeident.
	
	switch (step)
	{
	default:
	case 0:
		[self setCellDuration:QTMakeTimeWithTimeInterval(0.5)]; break;

	case 1:
		[self setCellDuration:QTMakeTimeWithTimeInterval(1.0)]; break;

	case 2:
		[self setCellDuration:QTMakeTimeWithTimeInterval(2.0)]; break;

	case 3:
		[self setCellDuration:QTMakeTimeWithTimeInterval(5.0)]; break;

	case 4:
		[self setCellDuration:QTMakeTimeWithTimeInterval(10.0)]; break;

	case 5:
		[self setCellDuration:QTMakeTimeWithTimeInterval(30.0)]; break;

	case 6:
		[self setCellDuration:QTMakeTimeWithTimeInterval(999999.0)]; break;
	}
}

- (QTTimeRange) selection { return mSelection; }
- (void) setSelection:(QTTimeRange) range
{
	static BOOL sInSetSelection = NO;
	
	if (!sInSetSelection)
	{
		sInSetSelection = YES;
		
		QTTime selectionEnd = QTTimeRangeEnd(mSelection);
		QTTime rangeEnd = QTTimeRangeEnd(range);
		BOOL selStartChanged = QTTimeCompare(range.time, mSelection.time) != NSOrderedSame;
		BOOL selEndChanged = QTTimeCompare(rangeEnd, selectionEnd) != NSOrderedSame;

		if (selStartChanged || selEndChanged)
		{
			unsigned oldSelStartIndex = [self _cellIndexForTime:mSelection.time];
			unsigned oldSelEndIndex = [self _cellIndexForTime:selectionEnd];
			unsigned newSelStartIndex = [self _cellIndexForTime:range.time];
			unsigned newSelEndIndex = [self _cellIndexForTime:rangeEnd];

			if (selStartChanged && selEndChanged)
				[self _refreshCellsForTimeRange:mSelection];
			else if (selStartChanged)
			{
				if (oldSelStartIndex != newSelStartIndex)
					[self _refreshCellsFrom:oldSelStartIndex to:newSelStartIndex];
				else
					[self _refreshCellForTime:mSelection.time];
			}
			else if (selEndChanged)
			{
				if (oldSelEndIndex != newSelEndIndex)
					[self _refreshCellsFrom:oldSelEndIndex to:newSelEndIndex];
				else
					[self _refreshCellForTime:selectionEnd];
			}

			mSelection = range;
			[mMovie setSelection:mSelection];
			[[self window] invalidateCursorRectsForView:self];

			if (selStartChanged && selEndChanged)
				[self _refreshCellsForTimeRange:mSelection];
			else if (selStartChanged)
			{
				if (oldSelStartIndex != newSelStartIndex)
					[self _refreshCellsFrom:oldSelStartIndex to:newSelStartIndex];
				else
					[self _refreshCellForTime:mSelection.time];
				[self _refreshCellForTime:selectionEnd];
			}
			else if (selEndChanged)
			{
				if (oldSelEndIndex != newSelEndIndex)
					[self _refreshCellsFrom:oldSelEndIndex to:newSelEndIndex];
				else
					[self _refreshCellForTime:selectionEnd];
			}
		}
	
		sInSetSelection = NO;
	}
}

- (QTTime) currentTime { return mCurrentTime; }
- (void) setCurrentTime:(QTTime) time
{
	static BOOL sInSetCurrentTime = NO;

	if (!sInSetCurrentTime)
	{
		time = [self _pinTime:time InRange:QTMakeTimeRange(QTZeroTime, mMovieDuration)];
	
		if (QTTimeCompare(time, mCurrentTime) != NSOrderedSame)
		{
			sInSetCurrentTime = YES;
			
			[self _refreshCellForTime:mCurrentTime];
			mCurrentTime = time;
			[mMovie setCurrentTime:time];
			[self _refreshCellForTime:mCurrentTime];
		}
		
		sInSetCurrentTime = NO;
	}
}

- (BOOL) highlightSelection { return mHighlightSelection; }
- (void) setHighlightSelection:(BOOL) newState
{
	if (mHighlightSelection != newState)
	{
		mHighlightSelection = newState;
		[self setNeedsDisplay:YES];
	}
}

- (void) setFrame:(NSRect) r
{
	if (!NSEqualRects(r, [self frame]))
	{
		[super setFrame:r];
		[self _configureForMovie];
	}
}

- (void) keyDown:(NSEvent*) event
{
	NSString* chars = [event characters];
	unichar ch = [chars length] == 1 ? [chars characterAtIndex:0] : 0;
	
	//	Space character toggles play/pause
	if (ch == ' ')
	{
		if ([self _movieIsPlaying])
			[mMovie stop];
		else
			[mMovie play];
	}
	else
		[super keyDown:event];
}

- (void) mouseDown:(NSEvent*) event
{
	NSPoint point = [self convertPoint:[event locationInWindow] fromView:nil];
	QTTime time = [self _timeForPoint:point];
	BOOL extendSelection = ([event modifierFlags] & NSShiftKeyMask) ? YES : NO;
	
	[mMovie stop];

	if (extendSelection && ![self _isZeroTimeRange:mSelection])
	{
		//	The user has shift-clicked - extend the current selection
		mDragKind = kLNSQTScrubDrag_AlteringSelection;
		[self _adjustSelectionStartForTime:time];

		[self mouseDragged:event];
	}
	else if (NSPointInRect(point, NSOffsetRect(NSInsetRect([self _frameForTime:mSelection.time], -3.0, 0.0), -kSelectionFrameInset, 0.0)))
	{
		//	User clicked in the left 'grabber' - extend the current selection
		mDragKind = kLNSQTScrubDrag_AlteringSelection;
		mSelectionStart = QTTimeRangeEnd(mSelection);
		[self mouseDragged:event];
	}
	else if (NSPointInRect(point, NSOffsetRect(NSInsetRect([self _frameForTime:QTTimeRangeEnd(mSelection)], -3.0, 0.0), kSelectionFrameInset, 0.0)))
	{
		//	User clicked in the right 'grabber' - extend the current selection
		mDragKind = kLNSQTScrubDrag_AlteringSelection;
		mSelectionStart = mSelection.time;
		[self mouseDragged:event];
	}
	else if (QTTimeInTimeRange(time, mSelection))
	{
		//	User clicked in the selected range of frames - start dragging the selected movie frames if the user
		//	actually moves the mouse...
		mDragKind = kLNSQTScrubDrag_BeginDraggingSelection;
	}
	else
	{
		//	User clicked outside the current selected range of frames - begin selecting a new range of movie frames
		mSelectionStart = time;
		mDragKind = kLNSQTScrubDrag_AlteringSelection;
		[self mouseDragged:event];
	}
	
	[self setCurrentTime:time];
}

- (void) mouseUp:(NSEvent*) event
{
	if (mDragKind == kLNSQTScrubDrag_AlteringSelection)
	{
		[self _refreshCellForTime:QTTimeRangeEnd(mSelection)]; // to remove selection duration
	}
	mDragKind = kLNSQTScrubDrag_None;
}

- (void) mouseDragged:(NSEvent*) event
{
	NSPoint point = [self convertPoint:[event locationInWindow] fromView:nil];

	switch (mDragKind)
	{
	case kLNSQTScrubDrag_AlteringSelection:
		if ([self _cellIndexForPoint:point strict:YES] != NSNotFound)
		{
			QTTime time = [self _timeForPoint:point];
			QTMovie* saveMovie = mMovie;
			
			[self setCurrentTime:time];
			
			mMovie = nil;
			[self _adjustSelectionForTime:time];	
			[self autoscroll:event];
			mMovie = saveMovie;
		}
		break;

	case kLNSQTScrubDrag_BeginDraggingSelection:
		{
			NSError* error = nil;
			QTMovie* subMovie = [mMovie movieWithTimeRange:mSelection error:&error];

			if (subMovie)
			{
				NSData* data = [subMovie movieFormatRepresentation];
				NSPasteboard* pboard = [NSPasteboard pasteboardWithName:NSDragPboard];
				NSImage* image = [[self _frameImageAtTime:mSelection.time canQueue:NO] scaledImageWithSize:mScaledMovieSize];
				NSSize imageSize = [image size];

				[pboard declareTypes:[NSArray arrayWithObject:QTMoviePasteboardType] owner:self];
				[pboard setData:data forType:QTMoviePasteboardType];

				[self dragImage:image
							 at:NSMakePoint(point.x - imageSize.width / 2.0, point.y + imageSize.height / 2.0)
						 offset:NSZeroSize
						  event:event
					 pasteboard:pboard
						 source:self
					  slideBack:YES];
			}
			
			mDragKind = kLNSQTScrubDrag_None;
		}
		break;
	}
}

- (void) mouseEntered:(NSEvent*) event
{
	[[self window] makeFirstResponder:self];
	[self mouseMoved:event];
}

- (void) mouseExited:(NSEvent*) event
{
}

- (void) mouseMoved:(NSEvent*) event
{
	NSPoint point = [self convertPoint:[event locationInWindow] fromView:nil];

	if (NSPointInRect(point, [self visibleRect]) && 
		![self _movieIsPlaying] && 
		[self _cellIndexForPoint:point strict:NO] != NSNotFound)
		[self setCurrentTime:[self _timeForPoint:point]];
}

- (IBAction) moveLeft:(id) sender
{
	BOOL extendSelection = ([[NSApp currentEvent] modifierFlags] & NSShiftKeyMask) ? YES : NO;
	BOOL gotoBeginning = ([[NSApp currentEvent] modifierFlags] & NSCommandKeyMask) ? YES : NO;
	
	if (gotoBeginning)
		[mMovie gotoBeginning];
	else
		[mMovie stepBackward];
	[self setCurrentTime:[mMovie currentTime]];

	if (extendSelection)
	{
		[self _adjustSelectionStartForTime:mCurrentTime];
		[self _adjustSelectionForTime:mCurrentTime];
	}
}

- (IBAction) moveRight:(id) sender
{
	BOOL extendSelection = ([[NSApp currentEvent] modifierFlags] & NSShiftKeyMask) ? YES : NO;
	BOOL gotoEnding = ([[NSApp currentEvent] modifierFlags] & NSCommandKeyMask) ? YES : NO;

	if (gotoEnding)
		[mMovie gotoEnd];
	else
		[mMovie stepForward];
	[self setCurrentTime:[mMovie currentTime]];

	if (extendSelection)
	{
		[self _adjustSelectionStartForTime:mCurrentTime];
		[self _adjustSelectionForTime:mCurrentTime];
	}
}

- (void) resetCursorRects
{
	if (mCells > 0)
	{
		//	Scrobbing area
		NSRect visibleRect = [self visibleRect];
		unsigned firstCell = [self _cellIndexForPoint:NSMakePoint(NSMinX(visibleRect), NSMinY(visibleRect)) strict:NO];
		unsigned lastCell = [self _cellIndexForPoint:NSMakePoint(NSMaxX(visibleRect), NSMaxY(visibleRect)) strict:NO];

		if (lastCell == NSNotFound)
			lastCell = mCells - 1;

		for (unsigned index = firstCell; index <= lastCell; ++index)
			[self addCursorRect:[self _frameForCellAtIndex:index] cursor:[NSCursor pointingHandCursor]];
		
		//	Selection grabber areas
		if (![self _isZeroTimeRange:mSelection])
		{
			[self addCursorRect:NSOffsetRect(NSInsetRect([self _frameForTime:mSelection.time], -3.0, 0.0), -kSelectionFrameInset, 0.0)
						 cursor:[NSCursor resizeLeftRightCursor]];
			[self addCursorRect:NSOffsetRect(NSInsetRect([self _frameForTime:QTTimeRangeEnd(mSelection)], -3.0, 0.0), kSelectionFrameInset, 0.0)
						 cursor:[NSCursor resizeLeftRightCursor]];
		}
	}
}

@end
